"""merge

Revision ID: 3f1dafab60b2
Revises: c092dabf3ee5, 6f5fe19894b7
Create Date: 2019-09-30 04:34:42.092031

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3f1dafab60b2'
down_revision = ('c092dabf3ee5', '6f5fe19894b7')
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""add-session_creation_id

Revision ID: 06184d82a211
Revises: 250e8656cf45
Create Date: 2020-12-24 19:58:44.515321

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '06184d82a211'
down_revision = '250e8656cf45'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('kernels', sa.Column('session_creation_id', sa.String(length=32), nullable=True))


def downgrade():
    op.drop_column('kernels', 'session_creation_id')



from collections import ChainMap
from datetime import datetime, timedelta
import functools
import hashlib, hmac
import logging
import secrets
from typing import (
    Any,
    Final,
    Iterable,
    Mapping,
    TYPE_CHECKING,
    Tuple,
    cast,
)

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
from aioredis import Redis
from aioredis.client import Pipeline as RedisPipeline
from dateutil.tz import tzutc
from dateutil.parser import parse as dtparse
import sqlalchemy as sa
import trafaret as t

from ai.backend.common import redis, validators as tx
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.plugin.hook import (
    ALL_COMPLETED,
    FIRST_COMPLETED,
    PASSED,
)

from ..models import (
    keypairs, keypair_resource_policies, users,
)
from ..models.user import UserRole, UserStatus, INACTIVE_USER_STATUSES, check_credential
from ..models.keypair import generate_keypair as _gen_keypair, generate_ssh_keypair
from ..models.group import association_groups_users, groups
from ..models.utils import execute_with_retry
from .exceptions import (
    AuthorizationFailed,
    GenericBadRequest,
    GenericForbidden,
    GenericNotFound,
    InternalServerError,
    InvalidAuthParameters,
    InvalidAPIParameters,
    RejectedByHook,
)
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, set_handler_attr, get_handler_attr

if TYPE_CHECKING:
    from .context import RootContext

log: Final = BraceStyleAdapter(logging.getLogger(__name__))

whois_timezone_info: Final = {
    "A": 1 * 3600,
    "ACDT": 10.5 * 3600,
    "ACST": 9.5 * 3600,
    "ACT": -5 * 3600,
    "ACWST": 8.75 * 3600,
    "ADT": 4 * 3600,
    "AEDT": 11 * 3600,
    "AEST": 10 * 3600,
    "AET": 10 * 3600,
    "AFT": 4.5 * 3600,
    "AKDT": -8 * 3600,
    "AKST": -9 * 3600,
    "ALMT": 6 * 3600,
    "AMST": -3 * 3600,
    "AMT": -4 * 3600,
    "ANAST": 12 * 3600,
    "ANAT": 12 * 3600,
    "AQTT": 5 * 3600,
    "ART": -3 * 3600,
    "AST": 3 * 3600,
    "AT": -4 * 3600,
    "AWDT": 9 * 3600,
    "AWST": 8 * 3600,
    "AZOST": 0 * 3600,
    "AZOT": -1 * 3600,
    "AZST": 5 * 3600,
    "AZT": 4 * 3600,
    "AoE": -12 * 3600,
    "B": 2 * 3600,
    "BNT": 8 * 3600,
    "BOT": -4 * 3600,
    "BRST": -2 * 3600,
    "BRT": -3 * 3600,
    "BST": 6 * 3600,
    "BTT": 6 * 3600,
    "C": 3 * 3600,
    "CAST": 8 * 3600,
    "CAT": 2 * 3600,
    "CCT": 6.5 * 3600,
    "CDT": -5 * 3600,
    "CEST": 2 * 3600,
    "CET": 1 * 3600,
    "CHADT": 13.75 * 3600,
    "CHAST": 12.75 * 3600,
    "CHOST": 9 * 3600,
    "CHOT": 8 * 3600,
    "CHUT": 10 * 3600,
    "CIDST": -4 * 3600,
    "CIST": -5 * 3600,
    "CKT": -10 * 3600,
    "CLST": -3 * 3600,
    "CLT": -4 * 3600,
    "COT": -5 * 3600,
    "CST": -6 * 3600,
    "CT": -6 * 3600,
    "CVT": -1 * 3600,
    "CXT": 7 * 3600,
    "ChST": 10 * 3600,
    "D": 4 * 3600,
    "DAVT": 7 * 3600,
    "DDUT": 10 * 3600,
    "E": 5 * 3600,
    "EASST": -5 * 3600,
    "EAST": -6 * 3600,
    "EAT": 3 * 3600,
    "ECT": -5 * 3600,
    "EDT": -4 * 3600,
    "EEST": 3 * 3600,
    "EET": 2 * 3600,
    "EGST": 0 * 3600,
    "EGT": -1 * 3600,
    "EST": -5 * 3600,
    "ET": -5 * 3600,
    "F": 6 * 3600,
    "FET": 3 * 3600,
    "FJST": 13 * 3600,
    "FJT": 12 * 3600,
    "FKST": -3 * 3600,
    "FKT": -4 * 3600,
    "FNT": -2 * 3600,
    "G": 7 * 3600,
    "GALT": -6 * 3600,
    "GAMT": -9 * 3600,
    "GET": 4 * 3600,
    "GFT": -3 * 3600,
    "GILT": 12 * 3600,
    "GMT": 0 * 3600,
    "GST": 4 * 3600,
    "GYT": -4 * 3600,
    "H": 8 * 3600,
    "HDT": -9 * 3600,
    "HKT": 8 * 3600,
    "HOVST": 8 * 3600,
    "HOVT": 7 * 3600,
    "HST": -10 * 3600,
    "I": 9 * 3600,
    "ICT": 7 * 3600,
    "IDT": 3 * 3600,
    "IOT": 6 * 3600,
    "IRDT": 4.5 * 3600,
    "IRKST": 9 * 3600,
    "IRKT": 8 * 3600,
    "IRST": 3.5 * 3600,
    "IST": 5.5 * 3600,
    "JST": 9 * 3600,
    "K": 10 * 3600,
    "KGT": 6 * 3600,
    "KOST": 11 * 3600,
    "KRAST": 8 * 3600,
    "KRAT": 7 * 3600,
    "KST": 9 * 3600,
    "KUYT": 4 * 3600,
    "L": 11 * 3600,
    "LHDT": 11 * 3600,
    "LHST": 10.5 * 3600,
    "LINT": 14 * 3600,
    "M": 12 * 3600,
    "MAGST": 12 * 3600,
    "MAGT": 11 * 3600,
    "MART": 9.5 * 3600,
    "MAWT": 5 * 3600,
    "MDT": -6 * 3600,
    "MHT": 12 * 3600,
    "MMT": 6.5 * 3600,
    "MSD": 4 * 3600,
    "MSK": 3 * 3600,
    "MST": -7 * 3600,
    "MT": -7 * 3600,
    "MUT": 4 * 3600,
    "MVT": 5 * 3600,
    "MYT": 8 * 3600,
    "N": -1 * 3600,
    "NCT": 11 * 3600,
    "NDT": 2.5 * 3600,
    "NFT": 11 * 3600,
    "NOVST": 7 * 3600,
    "NOVT": 7 * 3600,
    "NPT": 5.5 * 3600,
    "NRT": 12 * 3600,
    "NST": 3.5 * 3600,
    "NUT": -11 * 3600,
    "NZDT": 13 * 3600,
    "NZST": 12 * 3600,
    "O": -2 * 3600,
    "OMSST": 7 * 3600,
    "OMST": 6 * 3600,
    "ORAT": 5 * 3600,
    "P": -3 * 3600,
    "PDT": -7 * 3600,
    "PET": -5 * 3600,
    "PETST": 12 * 3600,
    "PETT": 12 * 3600,
    "PGT": 10 * 3600,
    "PHOT": 13 * 3600,
    "PHT": 8 * 3600,
    "PKT": 5 * 3600,
    "PMDT": -2 * 3600,
    "PMST": -3 * 3600,
    "PONT": 11 * 3600,
    "PST": -8 * 3600,
    "PT": -8 * 3600,
    "PWT": 9 * 3600,
    "PYST": -3 * 3600,
    "PYT": -4 * 3600,
    "Q": -4 * 3600,
    "QYZT": 6 * 3600,
    "R": -5 * 3600,
    "RET": 4 * 3600,
    "ROTT": -3 * 3600,
    "S": -6 * 3600,
    "SAKT": 11 * 3600,
    "SAMT": 4 * 3600,
    "SAST": 2 * 3600,
    "SBT": 11 * 3600,
    "SCT": 4 * 3600,
    "SGT": 8 * 3600,
    "SRET": 11 * 3600,
    "SRT": -3 * 3600,
    "SST": -11 * 3600,
    "SYOT": 3 * 3600,
    "T": -7 * 3600,
    "TAHT": -10 * 3600,
    "TFT": 5 * 3600,
    "TJT": 5 * 3600,
    "TKT": 13 * 3600,
    "TLT": 9 * 3600,
    "TMT": 5 * 3600,
    "TOST": 14 * 3600,
    "TOT": 13 * 3600,
    "TRT": 3 * 3600,
    "TVT": 12 * 3600,
    "U": -8 * 3600,
    "ULAST": 9 * 3600,
    "ULAT": 8 * 3600,
    "UTC": 0 * 3600,
    "UYST": -2 * 3600,
    "UYT": -3 * 3600,
    "UZT": 5 * 3600,
    "V": -9 * 3600,
    "VET": -4 * 3600,
    "VLAST": 11 * 3600,
    "VLAT": 10 * 3600,
    "VOST": 6 * 3600,
    "VUT": 11 * 3600,
    "W": -10 * 3600,
    "WAKT": 12 * 3600,
    "WARST": -3 * 3600,
    "WAST": 2 * 3600,
    "WAT": 1 * 3600,
    "WEST": 1 * 3600,
    "WET": 0 * 3600,
    "WFT": 12 * 3600,
    "WGST": -2 * 3600,
    "WGT": -3 * 3600,
    "WIB": 7 * 3600,
    "WIT": 9 * 3600,
    "WITA": 8 * 3600,
    "WST": 14 * 3600,
    "WT": 0 * 3600,
    "X": -11 * 3600,
    "Y": -12 * 3600,
    "YAKST": 10 * 3600,
    "YAKT": 9 * 3600,
    "YAPT": 10 * 3600,
    "YEKST": 6 * 3600,
    "YEKT": 5 * 3600,
    "Z": 0 * 3600,
}


def _extract_auth_params(request):
    """
    HTTP Authorization header must be formatted as:
    "Authorization: BackendAI signMethod=HMAC-SHA256,
                    credential=<ACCESS_KEY>:<SIGNATURE>"
    """
    auth_hdr = request.headers.get('Authorization')
    if not auth_hdr:
        return None
    pieces = auth_hdr.split(' ', 1)
    if len(pieces) != 2:
        raise InvalidAuthParameters('Malformed authorization header')
    auth_type, auth_str = pieces
    if auth_type not in ('BackendAI', 'Sorna'):
        raise InvalidAuthParameters('Invalid authorization type name')

    raw_params = map(lambda s: s.strip(), auth_str.split(','))
    params = {}
    for param in raw_params:
        key, value = param.split('=', 1)
        params[key.strip()] = value.strip()

    try:
        access_key, signature = params['credential'].split(':', 1)
        ret = params['signMethod'], access_key, signature
        return ret
    except (KeyError, ValueError):
        raise InvalidAuthParameters('Missing or malformed authorization parameters')


def check_date(request: web.Request) -> bool:
    raw_date = request.headers.get('Date')
    if not raw_date:
        raw_date = request.headers.get('X-BackendAI-Date',
                                       request.headers.get('X-Sorna-Date'))
    if not raw_date:
        return False
    try:
        # HTTP standard says "Date" header must be in GMT only.
        # However, dateutil.parser can recognize other commonly used
        # timezone names and offsets.
        date = dtparse(raw_date, tzinfos=whois_timezone_info)
        if date.tzinfo is None:
            date = date.replace(tzinfo=tzutc())  # assume as UTC
        now = datetime.now(tzutc())
        min_date = now - timedelta(minutes=15)
        max_date = now + timedelta(minutes=15)
        request['date'] = date
        request['raw_date'] = raw_date
        if not (min_date < date < max_date):
            return False
    except ValueError:
        return False
    return True


async def sign_request(sign_method: str, request: web.Request, secret_key: str) -> str:
    try:
        mac_type, hash_type = map(lambda s: s.lower(), sign_method.split('-'))
        assert mac_type == 'hmac', 'Unsupported request signing method (MAC type)'
        assert hash_type in hashlib.algorithms_guaranteed, \
               'Unsupported request signing method (hash type)'

        new_api_version = request.headers.get('X-BackendAI-Version')
        legacy_api_version = request.headers.get('X-Sorna-Version')
        api_version = new_api_version or legacy_api_version
        assert api_version is not None, 'API version missing in request headers'
        body = b''
        if api_version < 'v4.20181215':
            if (request.can_read_body and
                request.content_type != 'multipart/form-data'):
                # read the whole body if neither streaming nor bodyless
                body = await request.read()
        body_hash = hashlib.new(hash_type, body).hexdigest()

        sign_bytes = ('{0}\n{1}\n{2}\nhost:{3}\ncontent-type:{4}\n'
                      'x-{name}-version:{5}\n{6}').format(
            request.method, str(request.raw_path), request['raw_date'],
            request.host, request.content_type, api_version,
            body_hash,
            name='backendai' if new_api_version is not None else 'sorna'
        ).encode()
        sign_key = hmac.new(secret_key.encode(),
                            request['date'].strftime('%Y%m%d').encode(),
                            hash_type).digest()
        sign_key = hmac.new(sign_key, request.host.encode(), hash_type).digest()
        return hmac.new(sign_key, sign_bytes, hash_type).hexdigest()
    except ValueError:
        raise AuthorizationFailed('Invalid signature')
    except AssertionError as e:
        raise InvalidAuthParameters(e.args[0])


@web.middleware
async def auth_middleware(request: web.Request, handler) -> web.StreamResponse:
    """
    Fetches user information and sets up keypair, uesr, and is_authorized
    attributes.
    """
    # This is a global middleware: request.app is the root app.
    root_ctx: RootContext = request.app['_root.context']
    request['is_authorized'] = False
    request['is_admin'] = False
    request['is_superadmin'] = False
    request['keypair'] = None
    request['user'] = None
    if not get_handler_attr(request, 'auth_required', False):
        return (await handler(request))
    if not check_date(request):
        raise InvalidAuthParameters('Date/time sync error')

    # PRE_AUTH_MIDDLEWARE allows authentication via 3rd-party request headers/cookies.
    # Any responsible hook must return a valid keypair.
    hook_result = await root_ctx.hook_plugin_ctx.dispatch(
        'PRE_AUTH_MIDDLEWARE',
        (request,),
        return_when=FIRST_COMPLETED,
    )
    row = None
    if hook_result.status != PASSED:
        raise RejectedByHook.from_hook_result(hook_result)
    elif hook_result.result:
        # Passed one of the hook.
        # The "None" access_key means that the hook has allowed anonymous access.
        access_key = hook_result.result
        if access_key is not None:
            async def _query_cred():
                async with root_ctx.db.begin_readonly() as conn:
                    j = (
                        keypairs
                        .join(users, keypairs.c.user == users.c.uuid)
                        .join(
                            keypair_resource_policies,
                            keypairs.c.resource_policy == keypair_resource_policies.c.name,
                        )
                    )
                    query = (
                        sa.select([users, keypairs, keypair_resource_policies], use_labels=True)
                        .select_from(j)
                        .where(
                            (keypairs.c.access_key == access_key) &
                            (keypairs.c.is_active.is_(True))
                        )
                    )
                    result = await conn.execute(query)
                return result.first()

            row = await execute_with_retry(_query_cred)
            if row is None:
                raise AuthorizationFailed('Access key not found')

            async def _pipe_builder(r: Redis) -> RedisPipeline:
                pipe = r.pipeline()
                num_queries_key = f'kp:{access_key}:num_queries'
                pipe.incr(num_queries_key)
                pipe.expire(num_queries_key, 86400 * 30)  # retention: 1 month
                return pipe

            await redis.execute(root_ctx.redis_stat, _pipe_builder)
        else:
            # unsigned requests may be still accepted for public APIs
            pass
    else:
        # There were no hooks configured.
        # Perform our own authentication.
        params = _extract_auth_params(request)
        if params:
            sign_method, access_key, signature = params

            async def _query_cred():
                async with root_ctx.db.begin_readonly() as conn:
                    j = (
                        keypairs
                        .join(users, keypairs.c.user == users.c.uuid)
                        .join(keypair_resource_policies,
                              keypairs.c.resource_policy == keypair_resource_policies.c.name)
                    )
                    query = (
                        sa.select([users, keypairs, keypair_resource_policies], use_labels=True)
                        .select_from(j)
                        .where(
                            (keypairs.c.access_key == access_key) &
                            (keypairs.c.is_active.is_(True))
                        )
                    )
                    result = await conn.execute(query)
                    return result.first()

            row = await execute_with_retry(_query_cred)
            if row is None:
                raise AuthorizationFailed('Access key not found')
            my_signature = \
                await sign_request(sign_method, request, row['keypairs_secret_key'])
            if not secrets.compare_digest(my_signature, signature):
                raise AuthorizationFailed('Signature mismatch')

            async def _pipe_builder(r: Redis) -> RedisPipeline:
                pipe = r.pipeline()
                num_queries_key = f'kp:{access_key}:num_queries'
                pipe.incr(num_queries_key)
                pipe.expire(num_queries_key, 86400 * 30)  # retention: 1 month
                return pipe

            await redis.execute(root_ctx.redis_stat, _pipe_builder)
        else:
            # unsigned requests may be still accepted for public APIs
            pass

    if row is not None:
        auth_result = {
            'is_authorized': True,
            'keypair': {
                col.name: row[f'keypairs_{col.name}']
                for col in keypairs.c
                if col.name != 'secret_key'
            },
            'user': {
                col.name: row[f'users_{col.name}']
                for col in users.c
                if col.name not in ('password', 'description', 'created_at')
            },
            'is_admin': row['keypairs_is_admin'],
        }
        auth_result['keypair']['resource_policy'] = {
            col.name: row[f'keypair_resource_policies_{col.name}']
            for col in keypair_resource_policies.c
        }
        auth_result['user']['id'] = row['keypairs_user_id']  # legacy
        auth_result['is_superadmin'] = (auth_result['user']['role'] == 'superadmin')
        # Populate the result to the per-request state dict.
        request.update(auth_result)

    # No matter if authenticated or not, pass-through to the handler.
    # (if it's required, auth_required decorator will handle the situation.)
    return (await handler(request))


def auth_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


def admin_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False) and request.get('is_admin', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


def superadmin_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False) and request.get('is_superadmin', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('echo'): t.String,
    }))
async def test(request: web.Request, params: Any) -> web.Response:
    log.info('AUTH.TEST(ak:{})', request['keypair']['access_key'])
    resp_data = {'authorized': 'yes'}
    if 'echo' in params:
        resp_data['echo'] = params['echo']
    return web.json_response(resp_data)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('group', default=None): t.Null | tx.UUID,
    }))
async def get_role(request: web.Request, params: Any) -> web.Response:
    group_role = None
    root_ctx: RootContext = request.app['_root.context']
    log.info(
        'AUTH.ROLES(ak:{}, d:{}, g:{})',
        request['keypair']['access_key'],
        request['user']['domain_name'],
        params['group'],
    )
    if params['group'] is not None:
        query = (
            # TODO: per-group role is not yet implemented.
            sa.select([association_groups_users.c.group_id])
            .select_from(association_groups_users)
            .where(
                (association_groups_users.c.group_id == params['group']) &
                (association_groups_users.c.user_id == request['user']['uuid'])
            )
        )
        async with root_ctx.db.begin() as conn:
            result = await conn.execute(query)
            row = result.first()
            if row is None:
                raise GenericNotFound('No such user group or '
                                      'you are not the member of the group.')
        group_role = 'user'
    resp_data = {
        'global_role': 'superadmin' if request['is_superadmin'] else 'user',
        'domain_role': 'admin' if request['is_admin'] else 'user',
        'group_role': group_role,
    }
    return web.json_response(resp_data)


@atomic
@check_api_params(
    t.Dict({
        t.Key('type'): t.Enum('keypair', 'jwt'),
        t.Key('domain'): t.String,
        t.Key('username'): t.String,
        t.Key('password'): t.String,
    }))
async def authorize(request: web.Request, params: Any) -> web.Response:
    if params['type'] != 'keypair':
        # other types are not implemented yet.
        raise InvalidAPIParameters('Unsupported authorization type')
    log.info('AUTH.AUTHORIZE(d:{0[domain]}, u:{0[username]}, passwd:****, type:{0[type]})', params)
    root_ctx: RootContext = request.app['_root.context']

    # [Hooking point for AUTHORIZE with the FIRST_COMPLETED requirement]
    # The hook handlers should accept the whole ``params`` dict, and optional
    # ``db`` parameter (if the hook needs to query to database).
    # They should return a corresponding Backend.AI user object after performing
    # their own authentication steps, like LDAP authentication, etc.
    hook_result = await root_ctx.hook_plugin_ctx.dispatch(
        'AUTHORIZE',
        (request, params,),
        return_when=FIRST_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RejectedByHook.from_hook_result(hook_result)
    elif hook_result.result:
        # Passed one of AUTHORIZED hook
        user = hook_result.result
    else:
        # No AUTHORIZE hook is defined (proceed with normal login)
        user = await check_credential(
            root_ctx.db,
            params['domain'], params['username'], params['password']
        )
    if user is None:
        raise AuthorizationFailed('User credential mismatch.')
    if user['status'] == UserStatus.BEFORE_VERIFICATION:
        raise AuthorizationFailed('This account needs email verification.')
    if user['status'] in INACTIVE_USER_STATUSES:
        raise AuthorizationFailed('User credential mismatch.')
    async with root_ctx.db.begin() as conn:
        query = (sa.select([keypairs.c.access_key, keypairs.c.secret_key])
                   .select_from(keypairs)
                   .where(
                       (keypairs.c.user == user['uuid']) &
                       (keypairs.c.is_active)
                   )
                   .order_by(sa.desc(keypairs.c.is_admin)))
        result = await conn.execute(query)
        keypair = result.first()
    if keypair is None:
        raise AuthorizationFailed('No API keypairs found.')
    return web.json_response({
        'data': {
            'access_key': keypair['access_key'],
            'secret_key': keypair['secret_key'],
            'role': user['role'],
            'status': user['status'],
        },
    })


@atomic
@check_api_params(
    t.Dict({
        t.Key('domain'): t.String,
        t.Key('email'): t.String,
        t.Key('password'): t.String,
    }).allow_extra('*'))
async def signup(request: web.Request, params: Any) -> web.Response:
    log_fmt = 'AUTH.SIGNUP(d:{}, email:{}, passwd:****)'
    log_args = (params['domain'], params['email'])
    log.info(log_fmt, *log_args)
    root_ctx: RootContext = request.app['_root.context']

    # [Hooking point for PRE_SIGNUP with the ALL_COMPLETED requirement]
    # The hook handlers should accept the whole ``params`` dict.
    # They should return a dict to override the user information,
    # where the keys must be a valid field name of the users table,
    # with two exceptions: "resource_policy" (name) and "group" (name).
    # A plugin may return an empty dict if it has nothing to override.
    hook_result = await root_ctx.hook_plugin_ctx.dispatch(
        'PRE_SIGNUP',
        (params, ),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RejectedByHook.from_hook_result(hook_result)
    else:
        # Merge the hook results as a single map.
        user_data_overriden = ChainMap(*cast(Mapping, hook_result.result))

    async with root_ctx.db.begin() as conn:
        # Check if email already exists.
        query = (sa.select([users])
                   .select_from(users)
                   .where((users.c.email == params['email'])))
        result = await conn.execute(query)
        row = result.first()
        if row is not None:
            raise GenericBadRequest('Email already exists')

        # Create a user.
        data = {
            'domain_name': params['domain'],
            'username': params['username'] if 'username' in params else params['email'],
            'email': params['email'],
            'password': params['password'],
            'need_password_change': False,
            'full_name': params['full_name'] if 'full_name' in params else '',
            'description': params['description'] if 'description' in params else '',
            'status': UserStatus.ACTIVE,
            'status_info': 'user-signup',
            'role': UserRole.USER,
            'integration_id': None,
        }
        if user_data_overriden:
            for key, val in user_data_overriden.items():
                if key in data:  # take only valid fields
                    data[key] = val
        query = (users.insert().values(data))
        result = await conn.execute(query)
        if result.rowcount > 0:
            checkq = users.select().where(users.c.email == params['email'])
            result = await conn.execute(checkq)
            user = result.first()
            # Create user's first access_key and secret_key.
            ak, sk = _gen_keypair()
            resource_policy = (
                user_data_overriden.get('resource_policy', 'default')
            )
            kp_data = {
                'user_id': params['email'],
                'access_key': ak,
                'secret_key': sk,
                'is_active': True if data.get('status') == UserStatus.ACTIVE else False,
                'is_admin': False,
                'resource_policy': resource_policy,
                'concurrency_used': 0,
                'rate_limit': 1000,
                'num_queries': 0,
                'user': user.uuid,
            }
            query = (keypairs.insert().values(kp_data))
            await conn.execute(query)

            # Add user to the default group.
            group_name = user_data_overriden.get('group', 'default')
            query = (sa.select([groups.c.id])
                       .select_from(groups)
                       .where(groups.c.domain_name == params['domain'])
                       .where(groups.c.name == group_name))
            result = await conn.execute(query)
            grp = result.first()
            if grp is not None:
                values = [{'user_id': user.uuid, 'group_id': grp.id}]
                query = association_groups_users.insert().values(values)
                await conn.execute(query)
        else:
            raise InternalServerError('Error creating user account')

    resp_data = {
        'access_key': ak,
        'secret_key': sk,
    }

    # [Hooking point for POST_SIGNUP as one-way notification]
    # The hook handlers should accept a tuple of the user email,
    # the new user's UUID, and a dict with initial user's preferences.
    initial_user_prefs = {
        'lang': request.headers.get('Accept-Language', 'en-us').split(',')[0].lower(),
    }
    await root_ctx.hook_plugin_ctx.notify(
        'POST_SIGNUP',
        (params['email'], user.uuid, initial_user_prefs)
    )
    return web.json_response(resp_data, status=201)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['email', 'username']): t.String,
        t.Key('password'): t.String,
    }))
async def signout(request: web.Request, params: Any) -> web.Response:
    domain_name = request['user']['domain_name']
    log.info('AUTH.SIGNOUT(d:{}, email:{})', domain_name, params['email'])
    root_ctx: RootContext = request.app['_root.context']
    if request['user']['email'] != params['email']:
        raise GenericForbidden('Not the account owner')
    result = await check_credential(
        root_ctx.db,
        domain_name, params['email'], params['password'])
    if result is None:
        raise GenericBadRequest('Invalid email and/or password')
    async with root_ctx.db.begin() as conn:
        # Inactivate the user.
        query = (
            users.update()
            .values(status=UserStatus.INACTIVE)
            .where(users.c.email == params['email'])
        )
        await conn.execute(query)
        # Inactivate every keypairs of the user.
        query = (
            keypairs.update()
            .values(is_active=False)
            .where(keypairs.c.user_id == params['email'])
        )
        await conn.execute(query)
    return web.json_response({})


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('old_password'): t.String,
        t.Key('new_password'): t.String,
        t.Key('new_password2'): t.String,
    }))
async def update_password(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app['_root.context']
    domain_name = request['user']['domain_name']
    email = request['user']['email']
    log_fmt = 'AUTH.UDPATE_PASSWORD(d:{}, email:{})'
    log_args = (domain_name, email)
    log.info(log_fmt, *log_args)

    user = await check_credential(root_ctx.db, domain_name, email, params['old_password'])
    if user is None:
        log.info(log_fmt + ': old password mismtach', *log_args)
        raise AuthorizationFailed('Old password mismatch')
    if params['new_password'] != params['new_password2']:
        log.info(log_fmt + ': new password mismtach', *log_args)
        return web.json_response({'error_msg': 'new password mismitch'}, status=400)

    # [Hooking point for VERIFY_PASSWORD_FORMAT with the ALL_COMPLETED requirement]
    # The hook handlers should accept the old password and the new password and implement their
    # own password validation rules.
    # They should return None if the validation is successful and raise the Reject error
    # otherwise.
    hook_result = await root_ctx.hook_plugin_ctx.dispatch(
        'VERIFY_PASSWORD_FORMAT',
        (params['old_password'], params['new_password']),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        hook_result.reason = hook_result.reason or 'invalid password format'
        raise RejectedByHook.from_hook_result(hook_result)

    async with root_ctx.db.begin() as conn:
        # Update user password.
        data = {
            'password': params['new_password'],
            'need_password_change': False,
        }
        query = (users.update().values(data).where(users.c.email == email))
        await conn.execute(query)
    return web.json_response({}, status=200)


@atomic
@auth_required
async def get_ssh_keypair(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app['_root.context']
    domain_name = request['user']['domain_name']
    access_key = request['keypair']['access_key']
    log_fmt = 'AUTH.GET_SSH_KEYPAIR(d:{}, ak:{})'
    log_args = (domain_name, access_key)
    log.info(log_fmt, *log_args)
    async with root_ctx.db.begin() as conn:
        # Get SSH public key. Return partial string from the public key just for checking.
        query = (
            sa.select([keypairs.c.ssh_public_key])
            .where(keypairs.c.access_key == access_key)
        )
        pubkey = await conn.scalar(query)
    return web.json_response({'ssh_public_key': pubkey}, status=200)


@atomic
@auth_required
async def refresh_ssh_keypair(request: web.Request) -> web.Response:
    domain_name = request['user']['domain_name']
    access_key = request['keypair']['access_key']
    log_fmt = 'AUTH.REFRESH_SSH_KEYPAIR(d:{}, ak:{})'
    log_args = (domain_name, access_key)
    log.info(log_fmt, *log_args)
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        pubkey, privkey = generate_ssh_keypair()
        data = {
            'ssh_public_key': pubkey,
            'ssh_private_key': privkey,
        }
        query = (
            keypairs.update()
            .values(data)
            .where(keypairs.c.access_key == access_key)
        )
        await conn.execute(query)
    return web.json_response(data, status=200)


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app['prefix'] = 'auth'  # slashed to distinguish with "/vN/authorize"
    app['api_versions'] = (1, 2, 3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r''))
    cors.add(root_resource.add_route('GET', test))
    cors.add(root_resource.add_route('POST', test))
    test_resource = cors.add(app.router.add_resource('/test'))
    cors.add(test_resource.add_route('GET', test))
    cors.add(test_resource.add_route('POST', test))
    cors.add(app.router.add_route('POST', '/authorize', authorize))
    cors.add(app.router.add_route('GET', '/role', get_role))
    cors.add(app.router.add_route('POST', '/signup', signup))
    cors.add(app.router.add_route('POST', '/signout', signout))
    cors.add(app.router.add_route('POST', '/update-password', update_password))
    cors.add(app.router.add_route('GET', '/ssh-keypair', get_ssh_keypair))
    cors.add(app.router.add_route('PATCH', '/ssh-keypair', refresh_ssh_keypair))
    return app, [auth_middleware]



from __future__ import annotations

import enum
from typing import (
    Any,
    Mapping,
    Sequence,
    TYPE_CHECKING,
)

from dateutil.parser import parse as dtparse
import graphene
from graphene.types.datetime import DateTime as GQLDateTime
import sqlalchemy as sa
from sqlalchemy.sql.expression import true
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.engine.row import Row
from sqlalchemy.dialects import postgresql as pgsql

from ai.backend.common import msgpack, redis
from ai.backend.common.types import (
    AgentId,
    BinarySize,
    HardwareMetadata,
    ResourceSlot,
)

from .kernel import AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES, kernels
from .base import (
    metadata,
    batch_result,
    EnumType, Item, PaginatedList,
    ResourceSlotColumn,
)
from .minilang.queryfilter import QueryFilterParser
from .minilang.ordering import QueryOrderParser
if TYPE_CHECKING:
    from ai.backend.manager.models.gql import GraphQueryContext

__all__: Sequence[str] = (
    'agents', 'AgentStatus',
    'AgentList', 'Agent',
    'recalc_agent_resource_occupancy',
)


class AgentStatus(enum.Enum):
    ALIVE = 0
    LOST = 1
    RESTARTING = 2
    TERMINATED = 3


agents = sa.Table(
    'agents', metadata,
    sa.Column('id', sa.String(length=64), primary_key=True),
    sa.Column('status', EnumType(AgentStatus), nullable=False, index=True,
              default=AgentStatus.ALIVE),
    sa.Column('status_changed', sa.DateTime(timezone=True), nullable=True),
    sa.Column('region', sa.String(length=64), index=True, nullable=False),
    sa.Column('scaling_group', sa.ForeignKey('scaling_groups.name'), index=True,
              nullable=False, server_default='default', default='default'),
    sa.Column('schedulable', sa.Boolean(),
              nullable=False, server_default=true(), default=True),

    sa.Column('available_slots', ResourceSlotColumn(), nullable=False),
    sa.Column('occupied_slots', ResourceSlotColumn(), nullable=False),

    sa.Column('addr', sa.String(length=128), nullable=False),
    sa.Column('first_contact', sa.DateTime(timezone=True),
              server_default=sa.func.now()),
    sa.Column('lost_at', sa.DateTime(timezone=True), nullable=True),

    sa.Column('version', sa.String(length=64), nullable=False),
    sa.Column('compute_plugins', pgsql.JSONB(), nullable=False, default={}),
)


class Agent(graphene.ObjectType):

    class Meta:
        interfaces = (Item, )

    status = graphene.String()
    status_changed = GQLDateTime()
    region = graphene.String()
    scaling_group = graphene.String()
    schedulable = graphene.Boolean()
    available_slots = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    addr = graphene.String()
    first_contact = GQLDateTime()
    lost_at = GQLDateTime()
    live_stat = graphene.JSONString()
    version = graphene.String()
    compute_plugins = graphene.JSONString()
    hardware_metadata = graphene.JSONString()

    # Legacy fields
    mem_slots = graphene.Int()
    cpu_slots = graphene.Float()
    gpu_slots = graphene.Float()
    tpu_slots = graphene.Float()
    used_mem_slots = graphene.Int()
    used_cpu_slots = graphene.Float()
    used_gpu_slots = graphene.Float()
    used_tpu_slots = graphene.Float()
    cpu_cur_pct = graphene.Float()
    mem_cur_bytes = graphene.Float()

    compute_containers = graphene.List(
        'ai.backend.manager.models.ComputeContainer',
        status=graphene.String())

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> Agent:
        mega = 2 ** 20
        return cls(
            id=row['id'],
            status=row['status'].name,
            status_changed=row['status_changed'],
            region=row['region'],
            scaling_group=row['scaling_group'],
            schedulable=row['schedulable'],
            available_slots=row['available_slots'].to_json(),
            occupied_slots=row['occupied_slots'].to_json(),
            addr=row['addr'],
            first_contact=row['first_contact'],
            lost_at=row['lost_at'],
            version=row['version'],
            compute_plugins=row['compute_plugins'],
            # legacy fields
            mem_slots=BinarySize.from_str(row['available_slots']['mem']) // mega,
            cpu_slots=row['available_slots']['cpu'],
            gpu_slots=row['available_slots'].get('cuda.device', 0),
            tpu_slots=row['available_slots'].get('tpu.device', 0),
            used_mem_slots=BinarySize.from_str(
                row['occupied_slots'].get('mem', 0)) // mega,
            used_cpu_slots=float(row['occupied_slots'].get('cpu', 0)),
            used_gpu_slots=float(row['occupied_slots'].get('cuda.device', 0)),
            used_tpu_slots=float(row['occupied_slots'].get('tpu.device', 0)),
        )

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        rs = ctx.redis_stat
        live_stat = await redis.execute(rs, lambda r: r.get(str(self.id)))
        if live_stat is not None:
            live_stat = msgpack.unpackb(live_stat)
        return live_stat

    async def resolve_cpu_cur_pct(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        rs = ctx.redis_stat
        live_stat = await redis.execute(rs, lambda r: r.get(str(self.id)))
        if live_stat is not None:
            live_stat = msgpack.unpackb(live_stat)
            try:
                return float(live_stat['node']['cpu_util']['pct'])
            except (KeyError, TypeError, ValueError):
                return 0.0
        return 0.0

    async def resolve_mem_cur_bytes(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        rs = ctx.redis_stat
        live_stat = await redis.execute(rs, lambda r: r.get(str(self.id)))
        if live_stat is not None:
            live_stat = msgpack.unpackb(live_stat)
            try:
                return int(live_stat['node']['mem']['current'])
            except (KeyError, TypeError, ValueError):
                return 0
        return 0

    async def resolve_hardware_metadata(
        self,
        info: graphene.ResolveInfo,
    ) -> Mapping[str, HardwareMetadata]:
        graph_ctx: GraphQueryContext = info.context
        return await graph_ctx.registry.gather_agent_hwinfo(self.id)

    _queryfilter_fieldspec = {
        "id": ("id", None),
        "status": ("status", lambda s: AgentStatus[s]),
        "status_changed": ("status_changed", dtparse),
        "region": ("region", None),
        "scaling_group": ("scaling_group", None),
        "schedulable": ("schedulabe", None),
        "addr": ("addr", None),
        "first_contact": ("first_contat", dtparse),
        "lost_at": ("lost_at", dtparse),
        "version": ("version", None),
    }

    _queryorder_colmap = {
        "id": "id",
        "status": "status",
        "status_changed": "status_changed",
        "region": "region",
        "scaling_group": "scaling_group",
        "schedulable": "schedulable",
        "first_contact": "first_contact",
        "lost_at": "lost_at",
        "version": "version",
        "available_slots": "available_slots",
        "occupied_slots": "occupied_slots",
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext, *,
        scaling_group: str = None,
        raw_status: str = None,
        filter: str = None,
    ) -> int:
        query = (
            sa.select([sa.func.count(agents.c.id)])
            .select_from(agents)
        )
        if scaling_group is not None:
            query = query.where(agents.c.scaling_group == scaling_group)
        if raw_status is not None:
            query = query.where(agents.c.status == AgentStatus[raw_status])
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int, offset: int, *,
        scaling_group: str = None,
        raw_status: str = None,
        filter: str = None,
        order: str = None,
    ) -> Sequence[Agent]:
        query = (
            sa.select([agents])
            .select_from(agents)
            .limit(limit)
            .offset(offset)
        )
        if scaling_group is not None:
            query = query.where(agents.c.scaling_group == scaling_group)
        if raw_status is not None:
            query = query.where(agents.c.status == AgentStatus[raw_status])
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(
                agents.c.status.asc(),
                agents.c.scaling_group.asc(),
                agents.c.id.asc(),
            )
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                cls.from_row(graph_ctx, row)
                async for row in (await conn.stream(query))
            ]

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext, *,
        scaling_group: str = None,
        raw_status: str = None,
    ) -> Sequence[Agent]:
        query = (
            sa.select([agents])
            .select_from(agents)
        )
        if scaling_group is not None:
            query = query.where(agents.c.scaling_group == scaling_group)
        if raw_status is not None:
            query = query.where(agents.c.status == AgentStatus[raw_status])
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                cls.from_row(graph_ctx, row)
                async for row in (await conn.stream(query))
            ]

    @classmethod
    async def batch_load(
        cls,
        graph_ctx: GraphQueryContext,
        agent_ids: Sequence[AgentId], *,
        raw_status: str = None,
    ) -> Sequence[Agent | None]:
        query = (
            sa.select([agents])
            .select_from(agents)
            .where(agents.c.id.in_(agent_ids))
            .order_by(
                agents.c.id
            )
        )
        if raw_status is not None:
            query = query.where(agents.c.status == AgentStatus[raw_status])
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_result(
                graph_ctx, conn, query, cls,
                agent_ids, lambda row: row['id'],
            )


class AgentList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(Agent, required=True)


async def recalc_agent_resource_occupancy(db_conn: SAConnection, agent_id: AgentId) -> None:
    query = (
        sa.select([
            kernels.c.occupied_slots,
        ])
        .select_from(kernels)
        .where(
            (kernels.c.agent == agent_id) &
            (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
        )
    )
    occupied_slots = ResourceSlot()
    result = await db_conn.execute(query)
    for row in result:
        occupied_slots += row['occupied_slots']
    query = (
        sa.update(agents)
        .values({
            'occupied_slots': occupied_slots,
        })
        .where(agents.c.id == agent_id)
    )
    await db_conn.execute(query)



"""
Common definitions/constants used throughout the manager.
"""

import enum
import re
from typing import Final

from ai.backend.common.types import SlotName, SlotTypes


INTRINSIC_SLOTS: Final = {
    SlotName('cpu'): SlotTypes('count'),
    SlotName('mem'): SlotTypes('bytes'),
}


# The default container role name for multi-container sessions
DEFAULT_ROLE: Final = "main"

_RESERVED_VFOLDER_PATTERNS = [r'^\.[a-z0-9]+rc$', r'^\.[a-z0-9]+_profile$']
RESERVED_DOTFILES = ['.terminfo', '.jupyter', '.ssh', '.ssh/authorized_keys', '.local', '.config']
RESERVED_VFOLDERS = ['.terminfo', '.jupyter', '.tmux.conf', '.ssh']
RESERVED_VFOLDER_PATTERNS = [re.compile(x) for x in _RESERVED_VFOLDER_PATTERNS]

# Redis database IDs depending on purposes
REDIS_STAT_DB: Final = 0
REDIS_RLIM_DB: Final = 1
REDIS_LIVE_DB: Final = 2
REDIS_IMAGE_DB: Final = 3
REDIS_STREAM_DB: Final = 4


# PostgreSQL session-level advisory lock indentifiers
class AdvisoryLock(enum.IntEnum):
    LOCKID_TEST = 42
    LOCKID_SCHEDULE = 91
    LOCKID_PREPARE = 92
    LOCKID_SCHEDULE_TIMER = 191
    LOCKID_PREPARE_TIMER = 192
    LOCKID_LOG_CLEANUP_TIMER = 195
    LOCKID_IDLE_CHECK_TIMER = 196



import logging
from typing import AsyncIterator, Optional, cast

import aiohttp
import yarl

from ai.backend.common.logging import BraceStyleAdapter

from .base import BaseContainerRegistry

log = BraceStyleAdapter(logging.getLogger(__name__))


class HarborRegistry_v1(BaseContainerRegistry):

    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        api_url = self.registry_url / 'api'
        registry_projects = self.registry_info['project']
        rqst_args = {}
        if self.credentials:
            rqst_args['auth'] = aiohttp.BasicAuth(
                self.credentials['username'],
                self.credentials['password'],
            )
        project_list_url: Optional[yarl.URL]
        project_list_url = (api_url / 'projects').with_query(
            {'page_size': '30'}
        )
        project_ids = []
        while project_list_url is not None:
            async with sess.get(project_list_url, allow_redirects=False, **rqst_args) as resp:
                projects = await resp.json()
                for item in projects:
                    if item['name'] in registry_projects:
                        project_ids.append(item['project_id'])
                project_list_url = None
                next_page_link = resp.links.get('next')
                if next_page_link:
                    next_page_url = cast(yarl.URL, next_page_link['url'])
                    project_list_url = (
                        self.registry_url
                        .with_path(next_page_url.path)
                        .with_query(next_page_url.query)
                    )
        if not project_ids:
            log.warning('There is no given project.')
            return
        repo_list_url: Optional[yarl.URL]
        for project_id in project_ids:
            repo_list_url = (api_url / 'repositories').with_query(
                {'project_id': project_id, 'page_size': '30'}
            )
            while repo_list_url is not None:
                async with sess.get(repo_list_url, allow_redirects=False, **rqst_args) as resp:
                    items = await resp.json()
                    repos = [item['name'] for item in items]
                    for item in repos:
                        yield item
                    repo_list_url = None
                    next_page_link = resp.links.get('next')
                    if next_page_link:
                        next_page_url = cast(yarl.URL, next_page_link['url'])
                        repo_list_url = (
                            self.registry_url
                            .with_path(next_page_url.path)
                            .with_query(next_page_url.query)
                        )


class HarborRegistry_v2(BaseContainerRegistry):

    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        api_url = self.registry_url / 'api' / 'v2.0'
        registry_projects = self.registry_info['project']
        rqst_args = {}
        if self.credentials:
            rqst_args['auth'] = aiohttp.BasicAuth(
                self.credentials['username'],
                self.credentials['password'],
            )
        repo_list_url: Optional[yarl.URL]
        for project_name in registry_projects:
            repo_list_url = (api_url / 'projects' / project_name / 'repositories').with_query(
                {'page_size': '30'}
            )
            while repo_list_url is not None:
                async with sess.get(repo_list_url, allow_redirects=False, **rqst_args) as resp:
                    items = await resp.json()
                    if isinstance(items, dict) and (errors := items.get('errors', [])):
                        raise RuntimeError(f"failed to fetch repositories in project {project_name}",
                                           errors[0]['code'], errors[0]['message'])
                    repos = [item['name'] for item in items]
                    for item in repos:
                        yield item
                    repo_list_url = None
                    next_page_link = resp.links.get('next')
                    if next_page_link:
                        next_page_url = cast(yarl.URL, next_page_link['url'])
                        repo_list_url = (
                            self.registry_url
                            .with_path(next_page_url.path)
                            .with_query(next_page_url.query)
                        )



from __future__ import annotations

import asyncio
from typing import (
    Callable,
    Final,
    TYPE_CHECKING,
)

from .defs import AdvisoryLock

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
    from ai.backend.common.events import AbstractEvent, EventProducer


class GlobalTimer:

    """
    Executes the given async function only once in the given interval,
    uniquely among multiple manager instances across multiple nodes.
    """

    _event_producer: Final[EventProducer]

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        timer_id: int,
        event_producer: EventProducer,
        event_factory: Callable[[], AbstractEvent],
        interval: float = 10.0,
        initial_delay: float = 0.0,
    ) -> None:
        self.db = db
        self.timer_id = timer_id
        self._event_producer = event_producer
        self._event_factory = event_factory
        self.interval = interval
        self.initial_delay = initial_delay

    async def generate_tick(self) -> None:
        await asyncio.sleep(self.initial_delay)
        while True:
            async with self.db.advisory_lock(AdvisoryLock(self.timer_id)):
                await self._event_producer.produce_event(self._event_factory())
                await asyncio.sleep(self.interval)

    async def join(self) -> None:
        self._tick_task = asyncio.create_task(self.generate_tick())

    async def leave(self) -> None:
        if not self._tick_task.done():
            self._tick_task.cancel()
            try:
                await self._tick_task
            except asyncio.CancelledError:
                pass



"""keypairs_get_is_admin

Revision ID: 80176413d8aa
Revises: 4b8a66fb8d82
Create Date: 2017-09-14 16:01:59.994941

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql.expression import false

# revision identifiers, used by Alembic.
revision = '80176413d8aa'
down_revision = '4b8a66fb8d82'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('keypairs', sa.Column('is_admin', sa.Boolean(), nullable=False, default=False, server_default=false()))
    op.create_index(op.f('ix_keypairs_is_admin'), 'keypairs', ['is_admin'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_keypairs_is_admin'), table_name='keypairs')
    op.drop_column('keypairs', 'is_admin')
    # ### end Alembic commands ###



"""Add dotfile column to keypairs

Revision ID: 1e8531583e20
Revises: ce209920f654
Create Date: 2020-01-17 15:59:09.367691

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1e8531583e20'
down_revision = 'ce209920f654'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'keypairs',
        sa.Column('dotfiles', sa.LargeBinary(length=64 * 1024), nullable=False, server_default='\\x90')
    )


def downgrade():
    op.drop_column('keypairs', 'dotfiles')



import json
import logging
from typing import (
    Any,
    List,
    Mapping,
    TYPE_CHECKING,
    Tuple,
)
import uuid

from aiohttp import web
import aiohttp_cors
import sqlalchemy as sa
import trafaret as t
import yaml

from ai.backend.common import validators as tx
from ai.backend.common.logging import BraceStyleAdapter

from ..models import (
    association_groups_users as agus, domains,
    groups, session_templates, keypairs, users, UserRole,
    query_accessible_session_templates, TemplateType,
)
from ..models.session_template import check_task_template

from .auth import auth_required
from .exceptions import InvalidAPIParameters, TaskTemplateNotFound
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(t.Dict(
    {
        tx.AliasedKey(['group', 'groupName', 'group_name'], default='default'): t.String,
        tx.AliasedKey(['domain', 'domainName', 'domain_name'], default='default'): t.String,
        t.Key('owner_access_key', default=None): t.Null | t.String,
        t.Key('payload'): t.String
    }
))
async def create(request: web.Request, params: Any) -> web.Response:
    if params['domain'] is None:
        params['domain'] = request['user']['domain_name']
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    requester_uuid = request['user']['uuid']
    log.info(
        'SESSION_TEMPLATE.CREATE (ak:{0}/{1})',
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else '*',
    )
    user_uuid = request['user']['uuid']
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        if requester_access_key != owner_access_key:
            # Admin or superadmin is creating sessions for another user.
            # The check for admin privileges is already done in get_access_key_scope().
            query = (
                sa.select([keypairs.c.user, users.c.role, users.c.domain_name])
                .select_from(sa.join(keypairs, users, keypairs.c.user == users.c.uuid))
                .where(keypairs.c.access_key == owner_access_key)
            )
            result = await conn.execute(query)
            row = result.first()
            owner_domain = row['domain_name']
            owner_uuid = row['user']
            owner_role = row['role']
        else:
            # Normal case when the user is creating her/his own session.
            owner_domain = request['user']['domain_name']
            owner_uuid = requester_uuid
            owner_role = UserRole.USER

        query = (
            sa.select([domains.c.name])
            .select_from(domains)
            .where(
                (domains.c.name == owner_domain) &
                (domains.c.is_active)
            )
        )
        qresult = await conn.execute(query)
        domain_name = qresult.scalar()
        if domain_name is None:
            raise InvalidAPIParameters('Invalid domain')

        if owner_role == UserRole.SUPERADMIN:
            # superadmin can spawn container in any designated domain/group.
            query = (
                sa.select([groups.c.id])
                .select_from(groups)
                .where(
                    (groups.c.domain_name == params['domain']) &
                    (groups.c.name == params['group']) &
                    (groups.c.is_active)
                )
            )
            qresult = await conn.execute(query)
            group_id = qresult.scalar()
        elif owner_role == UserRole.ADMIN:
            # domain-admin can spawn container in any group in the same domain.
            if params['domain'] != owner_domain:
                raise InvalidAPIParameters("You can only set the domain to the owner's domain.")
            query = (
                sa.select([groups.c.id])
                .select_from(groups)
                .where(
                    (groups.c.domain_name == owner_domain) &
                    (groups.c.name == params['group']) &
                    (groups.c.is_active)
                )
            )
            qresult = await conn.execute(query)
            group_id = qresult.scalar()
        else:
            # normal users can spawn containers in their group and domain.
            if params['domain'] != owner_domain:
                raise InvalidAPIParameters("You can only set the domain to your domain.")
            query = (
                sa.select([agus.c.group_id])
                .select_from(agus.join(groups, agus.c.group_id == groups.c.id))
                .where(
                    (agus.c.user_id == owner_uuid) &
                    (groups.c.domain_name == owner_domain) &
                    (groups.c.name == params['group']) &
                    (groups.c.is_active)
                )
            )
            qresult = await conn.execute(query)
            group_id = qresult.scalar()
        if group_id is None:
            raise InvalidAPIParameters('Invalid group')

        log.debug('Params: {0}', params)
        try:
            body = json.loads(params['payload'])
        except json.JSONDecodeError:
            try:
                body = yaml.safe_load(params['payload'])
            except (yaml.YAMLError, yaml.MarkedYAMLError):
                raise InvalidAPIParameters('Malformed payload')
        template_data = check_task_template(body)
        template_id = uuid.uuid4().hex
        resp = {
            'id': template_id,
            'user': user_uuid.hex,
        }
        query = session_templates.insert().values({
            'id': template_id,
            'domain_name': params['domain'],
            'group_id': group_id,
            'user_uuid': user_uuid,
            'name': template_data['metadata']['name'],
            'template': template_data,
            'type': TemplateType.TASK,
        })
        result = await conn.execute(query)
        assert result.rowcount == 1
    return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('all', default=False): t.ToBool,
        tx.AliasedKey(['group_id', 'groupId'], default=None): tx.UUID | t.String | t.Null,
    }),
)
async def list_template(request: web.Request, params: Any) -> web.Response:
    resp = []
    access_key = request['keypair']['access_key']
    domain_name = request['user']['domain_name']
    user_role = request['user']['role']
    user_uuid = request['user']['uuid']
    log.info('SESSION_TEMPLATE.LIST (ak:{})', access_key)
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        entries: List[Mapping[str, Any]]
        if request['is_superadmin'] and params['all']:
            j = (
                session_templates
                .join(users, session_templates.c.user_uuid == users.c.uuid, isouter=True)
                .join(groups, session_templates.c.group_id == groups.c.id, isouter=True)
            )
            query = (
                sa.select([session_templates, users.c.email, groups.c.name], use_labels=True)
                .select_from(j)
                .where(
                    (session_templates.c.is_active) &
                    (session_templates.c.type == TemplateType.TASK)
                )
            )
            result = await conn.execute(query)
            entries = []
            for row in result:
                is_owner = True if row.session_templates_user == user_uuid else False
                entries.append({
                    'name': row.session_templates_name,
                    'id': row.session_templates_id,
                    'created_at': row.session_templates_created_at,
                    'is_owner': is_owner,
                    'user': (str(row.session_templates_user_uuid)
                             if row.session_templates_user_uuid else None),
                    'group': (str(row.session_templates_group_id)
                              if row.session_templates_group_id else None),
                    'user_email': row.users_email,
                    'group_name': row.groups_name,
                })
        else:
            extra_conds = None
            if params['group_id'] is not None:
                extra_conds = ((session_templates.c.group_id == params['group_id']))
            entries = await query_accessible_session_templates(
                conn,
                user_uuid,
                TemplateType.TASK,
                user_role=user_role,
                domain_name=domain_name,
                allowed_types=['user', 'group'],
                extra_conds=extra_conds,
            )
        for entry in entries:
            resp.append({
                'name': entry['name'],
                'id': entry['id'].hex,
                'created_at': str(entry['created_at']),
                'is_owner': entry['is_owner'],
                'user': str(entry['user']),
                'group': str(entry['group']),
                'user_email': entry['user_email'],
                'group_name': entry['group_name'],
                'type': 'user' if entry['user'] is not None else 'group',
            })
        return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('format', default='yaml'): t.Null | t.Enum('yaml', 'json'),
        t.Key('owner_access_key', default=None): t.Null | t.String,
    })
)
async def get(request: web.Request, params: Any) -> web.Response:
    if params['format'] not in ['yaml', 'json']:
        raise InvalidAPIParameters('format should be "yaml" or "json"')
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        'SESSION_TEMPLATE.GET (ak:{0}/{1})',
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else '*',
    )
    template_id = request.match_info['template_id']
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([session_templates.c.template])
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id) &
                (session_templates.c.is_active) &
                (session_templates.c.type == TemplateType.TASK)
            )
        )
        template = await conn.scalar(query)
        if not template:
            raise TaskTemplateNotFound
    template = json.loads(template)
    if params['format'] == 'yaml':
        body = yaml.dump(template)
        return web.Response(text=body, content_type='text/yaml')
    else:
        return web.json_response(template)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('payload'): t.String,
        t.Key('owner_access_key', default=None): t.Null | t.String,
    })
)
async def put(request: web.Request, params: Any) -> web.Response:
    template_id = request.match_info['template_id']

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        'SESSION_TEMPLATE.PUT (ak:{0}/{1})',
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else '*',
    )
    root_ctx: RootContext = request.app['_root.context']

    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([session_templates.c.id])
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id) &
                (session_templates.c.is_active) &
                (session_templates.c.type == TemplateType.TASK)
            )
        )
        result = await conn.scalar(query)
        if not result:
            raise TaskTemplateNotFound
        try:
            body = json.loads(params['payload'])
        except json.JSONDecodeError:
            body = yaml.safe_load(params['payload'])
        except (yaml.YAMLError, yaml.MarkedYAMLError):
            raise InvalidAPIParameters('Malformed payload')
        template_data = check_task_template(body)
        query = (
            sa.update(session_templates)
            .values(template=template_data, name=template_data['metadata']['name'])
            .where((session_templates.c.id == template_id))
        )
        result = await conn.execute(query)
        assert result.rowcount == 1

        return web.json_response({'success': True})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('owner_access_key', default=None): t.Null | t.String,
    })
)
async def delete(request: web.Request, params: Any) -> web.Response:
    template_id = request.match_info['template_id']
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        'SESSION_TEMPLATE.DELETE (ak:{0}/{1})',
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else '*'
    )
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([session_templates.c.id])
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id) &
                (session_templates.c.is_active) &
                (session_templates.c.type == TemplateType.TASK)
            )
        )
        result = await conn.scalar(query)
        if not result:
            raise TaskTemplateNotFound
        query = (
            sa.update(session_templates)
            .values(is_active=False)
            .where((session_templates.c.id == template_id))
        )
        result = await conn.execute(query)
        assert result.rowcount == 1

        return web.json_response({'success': True})


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['api_versions'] = (4, 5)
    app['prefix'] = 'template/session'
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('POST', '', create))
    cors.add(app.router.add_route('GET', '', list_template))
    template_resource = cors.add(app.router.add_resource(r'/{template_id}'))
    cors.add(template_resource.add_route('GET', get))
    cors.add(template_resource.add_route('PUT', put))
    cors.add(template_resource.add_route('DELETE', delete))

    return app, []



from __future__ import annotations

from typing import Sequence, TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from ai.backend.common.events import EventDispatcher, EventProducer
    from ai.backend.common.plugin.hook import HookPluginContext
    from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
    from ai.backend.common.types import RedisConnectionInfo

    from ..background import BackgroundTaskManager
    from ..models.storage import StorageSessionManager
    from ..models.utils import ExtendedAsyncSAEngine
    from ..idle import BaseIdleChecker
    from ..plugin.webapp import WebappPluginContext
    from ..registry import AgentRegistry
    from ..config import LocalConfig, SharedConfig
    from .types import CORSOptions


class BaseContext:
    pass


@attr.s(slots=True, auto_attribs=True, init=False)
class RootContext(BaseContext):
    pidx: int
    db: ExtendedAsyncSAEngine
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    redis_live: RedisConnectionInfo
    redis_stat: RedisConnectionInfo
    redis_image: RedisConnectionInfo
    redis_stream: RedisConnectionInfo
    shared_config: SharedConfig
    local_config: LocalConfig
    cors_options: CORSOptions

    webapp_plugin_ctx: WebappPluginContext
    idle_checkers: Sequence[BaseIdleChecker]
    storage_manager: StorageSessionManager
    hook_plugin_ctx: HookPluginContext

    registry: AgentRegistry

    error_monitor: ErrorPluginContext
    stats_monitor: StatsPluginContext
    background_task_manager: BackgroundTaskManager



from __future__ import annotations

import inspect
import logging
import re
from typing import (
    Any,
    Iterable,
    TYPE_CHECKING,
    Tuple,
)

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
import attr
import graphene
from graphql.execution.executors.asyncio import AsyncioExecutor
from graphql.execution import ExecutionResult
from graphql.error import GraphQLError, format_error
import trafaret as t

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common import validators as tx

from ..models.base import DataLoaderManager
from ..models.gql import (
    Mutations, Queries,
    GraphQueryContext,
    GQLMutationPrivilegeCheckMiddleware,
)
from .manager import GQLMutationUnfrozenRequiredMiddleware
from .exceptions import GraphQLError as BackendGQLError
from .auth import auth_required
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params
if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))

_rx_mutation_hdr = re.compile(r"^mutation(\s+\w+)?\s*(\(|{|@)", re.M)


class GQLLoggingMiddleware:

    def resolve(self, next, root, info: graphene.ResolveInfo, **args) -> Any:
        graph_ctx: GraphQueryContext = info.context
        if len(info.path) == 1:
            log.info('ADMIN.GQL (ak:{}, {}:{}, op:{})',
                     graph_ctx.access_key,
                     info.operation.operation,
                     info.field_name,
                     info.operation.name)
        return next(root, info, **args)


async def _handle_gql_common(request: web.Request, params: Any) -> ExecutionResult:
    root_ctx: RootContext = request.app['_root.context']
    app_ctx: PrivateContext = request.app['admin.context']
    manager_status = await root_ctx.shared_config.get_manager_status()
    known_slot_types = await root_ctx.shared_config.get_resource_slots()

    gql_ctx = GraphQueryContext(
        schema=app_ctx.gql_schema,
        dataloader_manager=DataLoaderManager(),
        local_config=root_ctx.local_config,
        shared_config=root_ctx.shared_config,
        etcd=root_ctx.shared_config.etcd,
        user=request['user'],
        access_key=request['keypair']['access_key'],
        db=root_ctx.db,
        redis_stat=root_ctx.redis_stat,
        redis_image=root_ctx.redis_image,
        manager_status=manager_status,
        known_slot_types=known_slot_types,
        background_task_manager=root_ctx.background_task_manager,
        storage_manager=root_ctx.storage_manager,
        registry=root_ctx.registry,
    )
    result = app_ctx.gql_schema.execute(
        params['query'],
        app_ctx.gql_executor,
        variable_values=params['variables'],
        operation_name=params['operation_name'],
        context_value=gql_ctx,
        middleware=[
            GQLLoggingMiddleware(),
            GQLMutationUnfrozenRequiredMiddleware(),
            GQLMutationPrivilegeCheckMiddleware(),
        ],
        return_promise=True)
    if inspect.isawaitable(result):
        result = await result
    return result


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('query'): t.String,
        t.Key('variables', default=None): t.Null | t.Mapping(t.String, t.Any),
        tx.AliasedKey(['operation_name', 'operationName'], default=None): t.Null | t.String,
    }))
async def handle_gql(request: web.Request, params: Any) -> web.Response:
    result = await _handle_gql_common(request, params)
    return web.json_response(result.to_dict(), status=200)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('query'): t.String,
        t.Key('variables', default=None): t.Null | t.Mapping(t.String, t.Any),
        tx.AliasedKey(['operation_name', 'operationName'], default=None): t.Null | t.String,
    }))
async def handle_gql_legacy(request: web.Request, params: Any) -> web.Response:
    # FIXME: remove in v21.09
    result = await _handle_gql_common(request, params)
    if result.errors:
        errors = []
        for e in result.errors:
            if isinstance(e, GraphQLError):
                errmsg = format_error(e)
                errors.append(errmsg)
            else:
                errmsg = {'message': str(e)}
                errors.append(errmsg)
            log.error('ADMIN.GQL Exception: {}', errmsg)
        raise BackendGQLError(extra_data=errors)
    return web.json_response(result.data, status=200)


@attr.s(auto_attribs=True, slots=True, init=False)
class PrivateContext:
    gql_executor: AsyncioExecutor
    gql_schema: graphene.Schema


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app['admin.context']
    app_ctx.gql_executor = AsyncioExecutor()
    app_ctx.gql_schema = graphene.Schema(
        query=Queries,
        mutation=Mutations,
        auto_camelcase=False,
    )


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['admin.context'] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('POST', r'/graphql', handle_gql_legacy))
    cors.add(app.router.add_route('POST', r'/gql', handle_gql))
    return app, []


if __name__ == '__main__':
    # If executed as a main program, print all GraphQL schemas.
    # (graphene transforms our object model into a textual representation)
    # This is useful for writing documentation!
    schema = graphene.Schema(
        query=Queries,
        mutation=Mutations,
        auto_camelcase=False)
    print('======== GraphQL API Schema ========')
    print(str(schema))



"""add tpu slot on kernel model

Revision ID: 352fa4f88f61
Revises: 57b523dec0e8
Create Date: 2018-11-12 11:39:30.613081

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '352fa4f88f61'
down_revision = '57b523dec0e8'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('kernels', sa.Column('tpu_set', sa.ARRAY(sa.Integer()), nullable=True))
    op.add_column('kernels', sa.Column('tpu_slot', sa.Float(), nullable=False,
                                       server_default='0'))
    op.alter_column('kernels', 'tpu_slot', server_default=None)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('kernels', 'tpu_slot')
    op.drop_column('kernels', 'tpu_set')
    # ### end Alembic commands ###
